@@ -30,7 +30,7 @@ from typing import Union
 
 import torch
 import torch.nn.functional as F
-from deprecated import deprecated
+from disent.util.deprecate import deprecated
 
 from disent.frameworks.helper.util import compute_ave_loss
 from disent.nn.modules import DisentModule
