@@ -39,7 +39,7 @@ def scale(r, a, b):
 
 def lerp(r, a, b):
     """Linear interpolation between parameters, respects bounds when t is out of bounds [0, 1]"""
-    assert a < b
+    # assert a < b
     r = np.clip(r, 0., 1.)
     # precise method, guarantees v==b when t==1 | simplifies to: a + t*(b-a)
     return (1 - r) * a + r * b
